@@ -9,6 +9,7 @@
 
 #include <KColorScheme>
 #include <KPluginFactory>
+#include <KWindowSystem>
 
 #include <QDBusConnection>
 #include <QDBusMessage>
@@ -23,6 +24,7 @@
 #include "config_editor/gtk2.h"
 #include "config_editor/settings_ini.h"
 #include "config_editor/xsettings.h"
+#include "gsd-xsettings-manager/gsd-xsettings-manager.h"
 
 K_PLUGIN_CLASS_WITH_JSON(GtkConfig, "gtkconfig.json")
 
@@ -32,6 +34,7 @@ GtkConfig::GtkConfig(QObject *parent, const QVariantList &)
     , themePreviewer(new ThemePreviewer(this))
     , kdeglobalsConfigWatcher(KConfigWatcher::create(KSharedConfig::openConfig()))
     , kwinConfigWatcher(KConfigWatcher::create(KSharedConfig::openConfig(QStringLiteral("kwinrc"))))
+    , kcmfontsConfigWatcher(KConfigWatcher::create(KSharedConfig::openConfig(QStringLiteral("kcmfonts"))))
     , kcminputConfigWatcher(KConfigWatcher::create(KSharedConfig::openConfig(QStringLiteral("kcminputrc"))))
     , breezeConfigWatcher(KConfigWatcher::create(KSharedConfig::openConfig(QStringLiteral("breezerc"))))
 {
@@ -39,8 +42,13 @@ GtkConfig::GtkConfig(QObject *parent, const QVariantList &)
     dbus.registerService(QStringLiteral("org.kde.GtkConfig"));
     dbus.registerObject(QStringLiteral("/GtkConfig"), this, QDBusConnection::ExportScriptableSlots);
 
+    if (qgetenv("GTK_USE_PORTAL") != "1" && KWindowSystem::isPlatformWayland()) {
+        m_gsdXsettingsManager = new GSDXSettingsManager(this);
+    }
+
     connect(kdeglobalsConfigWatcher.data(), &KConfigWatcher::configChanged, this, &GtkConfig::onKdeglobalsSettingsChange);
     connect(kwinConfigWatcher.data(), &KConfigWatcher::configChanged, this, &GtkConfig::onKWinSettingsChange);
+    connect(kcmfontsConfigWatcher.data(), &KConfigWatcher::configChanged, this, &GtkConfig::onKCMFontsSettingsChange);
     connect(kcminputConfigWatcher.data(), &KConfigWatcher::configChanged, this, &GtkConfig::onKCMInputSettingsChange);
     connect(breezeConfigWatcher.data(), &KConfigWatcher::configChanged, this, &GtkConfig::onBreezeSettingsChange);
 
@@ -152,6 +160,10 @@ void GtkConfig::setDarkThemePreference() const
 {
     const bool preferDarkTheme = configValueProvider->preferDarkTheme();
     SettingsIniEditor::setValue(QStringLiteral("gtk-application-prefer-dark-theme"), preferDarkTheme);
+    // https://gitlab.gnome.org/GNOME/gsettings-desktop-schemas/-/blob/master/headers/gdesktop-enums.h
+    GSettingsEditor::setValueAsEnum(QStringLiteral("color-scheme"),
+                                    preferDarkTheme ? 1 /*G_DESKTOP_COLOR_SCHEME_PREFER_DARK*/ : 0 /*G_DESKTOP_COLOR_SCHEME_DEFAULT*/,
+                                    QStringLiteral("org.gnome.desktop.interface"));
 }
 
 void GtkConfig::setWindowDecorationsAppearance() const
@@ -179,6 +191,9 @@ void GtkConfig::setEnableAnimations() const
     GSettingsEditor::setValue(QStringLiteral("enable-animations"), enableAnimations);
     SettingsIniEditor::setValue(QStringLiteral("gtk-enable-animations"), enableAnimations);
     XSettingsEditor::setValue(QStringLiteral("Gtk/EnableAnimations"), enableAnimations);
+    if (m_gsdXsettingsManager) {
+        m_gsdXsettingsManager->enableAnimationsChanged();
+    }
 }
 
 void GtkConfig::setGlobalScale() const
@@ -189,24 +204,40 @@ void GtkConfig::setGlobalScale() const
 
 void GtkConfig::setTextScale() const
 {
-    constexpr int baseTextDpi = 96 * 1024;
-    const double scaleFactor = configValueProvider->x11GlobalScaleFactor();
-    const double fractionalPart = std::fmod(scaleFactor, 1.0);
-    const double integerPart = std::max(scaleFactor - fractionalPart, 1.0);
-    const int textScaleAbsolute = baseTextDpi * scaleFactor;
-    const int textScaleRelative = baseTextDpi * (1.0 + fractionalPart / integerPart);
+    const double x11Scale = configValueProvider->x11GlobalScaleFactor();
+    const int x11ScaleIntegerPart = int(x11Scale);
 
-    SettingsIniEditor::setValue(QStringLiteral("gtk-xft-dpi"), textScaleAbsolute);
-    // GTK2
-    XSettingsEditor::setValue(QStringLiteral("Xft/DPI"), textScaleAbsolute);
-    // GTK3 and newer
-    XSettingsEditor::setValue(QStringLiteral("Gdk/UnscaledDPI"), textScaleRelative);
+    const int forceFontDpi = configValueProvider->fontDpi();
+
+    int x11TextDpiAbsolute = 96 * 1024;
+    double waylandTextScaleFactor = 1.0;
+
+    if (forceFontDpi == 0) {
+        x11TextDpiAbsolute = (96 * 1024) * x11Scale;
+    } else {
+        x11TextDpiAbsolute = (forceFontDpi * 1024);
+
+        if (!KWindowSystem::isPlatformX11()) {
+            x11TextDpiAbsolute *= x11Scale;
+        }
+
+        waylandTextScaleFactor = double(forceFontDpi) / 96.0;
+        waylandTextScaleFactor = std::clamp(waylandTextScaleFactor, 0.5, 3.0);
+    }
+
+    XSettingsEditor::unsetValue(QStringLiteral("Xft/DPI"));
+    SettingsIniEditor::setValue(QStringLiteral("gtk-xft-dpi"), x11TextDpiAbsolute);
+    XSettingsEditor::setValue(QStringLiteral("Gdk/UnscaledDPI"), x11TextDpiAbsolute / x11ScaleIntegerPart);
+    GSettingsEditor::setValue(QStringLiteral("text-scaling-factor"), waylandTextScaleFactor);
 }
 
 void GtkConfig::setColors() const
 {
     const QMap<QString, QColor> colors = configValueProvider->colors();
     CustomCssEditor::setColors(colors);
+    if (m_gsdXsettingsManager) {
+        m_gsdXsettingsManager->modulesChanged();
+    }
 }
 
 void GtkConfig::applyAllSettings() const
@@ -286,6 +317,15 @@ void GtkConfig::onKWinSettingsChange(const KConfigGroup &group, const QByteArray
     }
 }
 
+void GtkConfig::onKCMFontsSettingsChange(const KConfigGroup &group, const QByteArrayList &names) const
+{
+    if (group.name() == QStringLiteral("General")) {
+        if (names.contains("forceFontDPI") || names.contains("forceFontDPIWayland")) {
+            setTextScale();
+        }
+    }
+}
+
 void GtkConfig::onKCMInputSettingsChange(const KConfigGroup &group, const QByteArrayList &names) const
 {
     if (group.name() == QStringLiteral("Mouse")) {
